/** This file is part of Domotz Agent.
 * Copyright (C) 2016  Domotz Ltd
 *
 * Domotz Agent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Domotz Agent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Domotz Agent.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Created by Iacopo Papalini <iacopo@domotz.com>
 */

const vm = require('vm');
const logger = require('./library/logger').logger;
const util = require('util');

var myLog = '';

var newConsole = null;

var startTime = null;
var agentDriverSettings = null;

function purge_data_if_too_big(data, newConsole) {
    var maxDataSize = agentDriverSettings.max_data_len;
    const dataLength = Buffer(JSON.stringify(data)).length;
    if (dataLength > maxDataSize) {
        newConsole.warning("Trimming data to zero byte since it's too long: %s > %s",
            dataLength,
            maxDataSize);
        return null;
    } else {
        return data;
    }
}

function failure(data) {
    process.send({
        outcome: 'failure',
        data: purge_data_if_too_big(data, newConsole),
        log: newConsole.get(),
        myLog: myLog,
        elapsed: new Date() - startTime
    });
}

function success(data, variables) {
    var maxVarPerDevice = agentDriverSettings.max_var_per_device;
    if (variables && variables.length > maxVarPerDevice) {
        newConsole.warning("Too many declared variables (%s) - slicing to %s",
            variables.length,
            maxVarPerDevice);
        variables = variables.slice(0, maxVarPerDevice);
    }
    process.send({
        outcome: 'success',
        data: purge_data_if_too_big(data, newConsole),
        variables: variables,
        log: newConsole.get(),
        myLog: myLog.toString(),
        elapsed: new Date() - startTime
    });
}

function timeoutReached() {
    myLog += "Timeout expired\n";
    failure({'message': 'Timeout expired'});
}


process.on('message', function (message) {
    agentDriverSettings = message.agentDriverSettings;
    newConsole = logger(message.logLevel || 'info', agentDriverSettings.max_log_entries);
    var context = vm.createContext({
        D: {
            htmlParse: require('cheerio').load,
            success: success,
            failure: failure,
            math: {
                percent: function (actual, maximum) {
                    return Math.round(10000.0 * parseInt(actual, 10) / (parseInt(maximum, 10))) / 100;
                }
            },
            _: require('lodash'),
            q: require('q')
        },
        import: function () {
            failure({'message': "Import not possible in sandbox"});
        },
        require: function () {
            failure({'message': "Require not possible in sandbox"});
        },
        console: newConsole
    });

    if (message.device) {
        context.D.device = require('./library/device').device(message.device, message.driverName, agentDriverSettings, newConsole);
        newConsole.debug("Created new device for: %s for driver: %s", util.inspect(message.device), util.inspect(message.driverName));
    }

    const script = new vm.Script(message.script);
    var timeout = message.timeout || 5000;
    setTimeout(timeoutReached, timeout);

    startTime = new Date();
    script.runInContext(context, {
        displayErrors: true,
        fileName: 'custom script',
        lineOffset: 0
    });
});

process.on('uncaughtException', function (err) {
    myLog += 'Caught exception: ' + err.stack;
    failure({message: "Unhandled Exception in script", error: err.toString(), trace: err.stack});
});